/* eslint-disable no-console */
import type { LanguageRegistration, ThemedToken, ThemeRegistrationAny } from '@shikijs/types'
import { splitTokens } from '@shikijs/core'
import React, { type FunctionComponent } from 'react'
import { ActivityIndicator, ScrollView, type ScrollViewProps, type StyleProp, Text, type TextStyle, View, type ViewStyle } from 'react-native'
import php from 'shiki/dist/langs/php.mjs'
import tsx from 'shiki/dist/langs/tsx.mjs'
import typescript from 'shiki/dist/langs/typescript.mjs'
import { getHighlighter } from './SyntaxContext'

const LANGS = {
  php,
  tsx,
  typescript,
  // Add more languages as needed
} as const

const DEBUG = true
function log(message: string, ...args: any[]) {
  if (DEBUG) {
    console.log(`[SyntaxHighlighter] ${message}`, ...args)
  }
}

function extractPatternsFromGrammar(grammar: LanguageRegistration) {
  const repository = grammar?.repository || {}
  const breakpoints = new Set<number>()

  function processPatterns(patterns: any[]) {
    patterns.forEach((pattern) => {
      // For includes, fetch from repository
      if (pattern.include) {
        const key = pattern.include.replace('#', '')
        if (repository[key]) {
          if (repository[key].patterns)
            processPatterns(repository[key].patterns)

          if (repository[key].begin)
            breakpoints.add(0)

          if (repository[key].match)
            breakpoints.add(repository[key].match.length || 0)
        }
      }
      // Handle direct patterns
      else {
        if (pattern.begin) {
          breakpoints.add(0)
          if (pattern.end)
            breakpoints.add(pattern.end.length || 0)
        }
        if (pattern.match)
          breakpoints.add(pattern.match.length || 0)

        // Process nested patterns
        if (pattern.patterns)
          processPatterns(pattern.patterns)
      }
    })
  }

  // Start with top-level patterns
  if (grammar.patterns)
    processPatterns(grammar.patterns)

  return breakpoints
}

function processTokens(tokens: ThemedToken[][], language: string): ThemedToken[][] {
  // Get the language grammar
  const grammar = LANGS[language as keyof typeof LANGS]
  if (!grammar) {
    log('No grammar found for language:', language)
    return tokens
  }

  // @ts-expect-error - TS doesn't know about the `repository` property
  const grammarBreakpoints = extractPatternsFromGrammar(language)

  // Add token boundary breakpoints
  tokens.forEach((line) => {
    line.forEach((token) => {
      const start = token.offset || 0
      const end = start + token.content.length
      grammarBreakpoints.add(start)
      grammarBreakpoints.add(end)

      // Add breakpoints for specific syntax patterns within the token
      // This helps catch sub-token patterns like operators within identifiers
      if (token.content.length > 1) {
        token.content.split(/([(){}[\]<>.,;:?=+\-*/%&|^!~`@#$]+)/).forEach((part, index, parts) => {
          if (part) {
            const partStart = start + parts.slice(0, index).join('').length
            grammarBreakpoints.add(partStart)
            grammarBreakpoints.add(partStart + part.length)
          }
        })
      }
    })
  })

  // Use Shiki's splitTokens with our collected breakpoints
  return splitTokens(tokens, grammarBreakpoints)
}

export interface SyntaxHighlighterProps {
  text: string
  language: string
  theme: ThemeRegistrationAny
  fontSize?: number
  textStyle?: StyleProp<TextStyle>
  scrollViewProps?: ScrollViewProps
  containerStyle?: StyleProp<ViewStyle>
  highlightLines?: number[]
}

const defaultHighlightLines: number[] = []

export const SyntaxHighlighter: FunctionComponent<SyntaxHighlighterProps> = ({
  text,
  language,
  theme,
  fontSize = 14,
  textStyle,
  scrollViewProps,
  containerStyle,
  highlightLines = defaultHighlightLines,
}) => {
  const [loading, setLoading] = React.useState(true)
  const [error, setError] = React.useState<string | null>(null)
  const [tokens, setTokens] = React.useState<ThemedToken[][]>([])

  React.useEffect(() => {
    const highlight = async () => {
      try {
        log('Starting highlighting process')
        log('Input:', { text, language, theme: theme.name })

        setLoading(true)
        const highlighter = await getHighlighter()
        log('Highlighter instance obtained')

        if (!text || !language || !theme) {
          throw new Error('Missing required props')
        }

        log('Attempting tokenization...')
        const result = highlighter.codeToTokens(text, {
          lang: language,
          theme,
          includeExplanation: true,
        })

        log('Raw tokenization result:', result)

        if (!result?.tokens) {
          throw new Error('Invalid tokenization result')
        }

        // Process tokens using the loaded grammar
        const processedTokens = await processTokens(result.tokens, language)
        log('Processed tokens:', processedTokens)

        setTokens(processedTokens)
        setError(null)
      }
      catch (err) {
        const errorMessage = err instanceof Error ? err.message : 'Failed to highlight code'
        log('Error during highlighting:', err)
        setError(errorMessage)
        setTokens([])
      }
      finally {
        setLoading(false)
      }
    }

    highlight()
  }, [text, language, theme])

  const renderToken = (token: ThemedToken, index: number) => {
    return (
      <Text
        key={`${token.offset || index}-${index}`}
        style={[
          {
            color: token.color,
            fontSize,
            fontFamily: 'monospace',
            fontStyle: token.fontStyle === 2 ? 'italic' : 'normal',
            fontWeight: token.fontStyle === 1 ? 'bold' : 'normal',
          },
          textStyle,
        ]}
      >
        {token.content}
      </Text>
    )
  }

  const renderLine = (lineTokens: ThemedToken[], index: number) => {
    const isHighlighted = highlightLines.includes(index + 1)
    return (
      <View
        key={index}
        style={[
          {
            flexDirection: 'row',
            flexWrap: 'wrap',
            backgroundColor: isHighlighted ? 'rgba(255,255,255,0.1)' : 'transparent',
          },
        ]}
      >
        {lineTokens.map((token, tokenIndex) => renderToken(token, tokenIndex))}
      </View>
    )
  }

  if (loading) {
    return (
      <View style={[{ padding: 10 }, containerStyle]}>
        <ActivityIndicator />
      </View>
    )
  }

  if (error) {
    return (
      <View style={[{ padding: 10 }, containerStyle]}>
        <Text style={{ color: '#ff6b6b' }}>
          Error:
          {error}
        </Text>
      </View>
    )
  }

  return (
    <ScrollView {...scrollViewProps} horizontal style={[{ backgroundColor: theme.colors?.background || '#1e1e1e' }, scrollViewProps?.style]} contentContainerStyle={[{ padding: 10 }, containerStyle]}>
      <View onStartShouldSetResponder={() => true}>{tokens.map((lineTokens, index) => renderLine(lineTokens, index))}</View>
    </ScrollView>
  )
}
